// Ao40Fec.cpp: implementation of the CAo40Fec class.
//
//////////////////////////////////////////////////////////////////////
// Moe Wheatley AE4JY  <ae4jy@qsl.net>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
////////////////////////////////////////////////////////////////////////
//
//  This program is based on material largely developed by Phil Karn KA9Q
//  over the period 1997-2002.  Much of it can be found in libraries on 
//  http://www.ka9q.net/ao40/  
//				Copyright 1997 Phil Karn, KA9Q
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Ao40Fec.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//Reed Solomon Encoder constants
#define MM 8
#define NN 255
#define NROOTS 32
#define FCR 112
#define PRIM 11
#define IPRIM 116

//Viterbi Encoder constants
#define	V27POLYA	0x6d
#define	V27POLYB	0x4f

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAo40Fec::CAo40Fec()
{
	m_RS.alpha_to = NULL;
	m_RS.genpoly = NULL;
	m_RS.index_of = NULL;
	Init();
}

CAo40Fec::~CAo40Fec()
{
	if(m_RS.alpha_to)
		delete m_RS.alpha_to;
	if(m_RS.genpoly)
		delete m_RS.genpoly;
	if(m_RS.index_of)
		delete m_RS.index_of;
}

void CAo40Fec::Init()
{
int sr,i,j;

	//Generate Sync Vector
	sr = 0x7f;
	for(i=0; i<COLUMNS; i++)
	{
		m_SyncVector[i] = (sr & 64) ? 1 : 0;
		sr = (sr << 1) | parityb(sr & PNPOLY);
	}
	//Generate Scrambler Sequence
	sr = 0xff;
	for(i=0; i<sizeof(m_Scrambler); i++)
	{
		m_Scrambler[i] = sr;
		for(j=0; j<8; j++)
			sr = (sr << 1) | parity(sr & SCRAMBLER_POLY);
	}
	InitRS(8,0x187,112,11,32,RSPAD); // CCSDS standard
}

//////////////////////////////////////////////////////////////////////////
//Encode a frame. The input bytes are read from data[], and the output
//symbols, in transmit order, are placed in output[]
////////////////////////////////////////////////////////////////////
//				Copyright 1997 Phil Karn, KA9Q
////////////////////////////////////////////////////////////////////
void CAo40Fec::EncodeFrame(int* pOutput, unsigned char* pData)
{
int w,sr,si,row,col,i;
unsigned char rsblocks[RSBLOCKS][RSTOT];	//RSBLOCKS==2 RSTOT==160

	//Interleave data into Reed-Solomon codewords
	i=0;
	for(col=0;col<RSDATA;col++)	//RSDATA==128
	{
		for(row=0;row<RSBLOCKS;row++)	//RSBLOCKS==2
		{
			rsblocks[row][col] = pData[i++];
		}
	}

	//Encode RS codewords
	for(i=0;i<RSBLOCKS;i++)	//RSBLOCKS==2
		EncodeRS(&rsblocks[i][0],&rsblocks[i][RSDATA],RSPAD);

	//Generate sync vector
	w = 0;
	for(i=0;i<COLUMNS;i++)	//COLUMNS==65
	{
		pOutput[w] = m_SyncVector[i];
		w = update(w);
	}
  
	sr = 0; /* Initialize convolutional encoder */
	si = 0; /* Init scrambler */
	for(col=0; col<RSTOT; col++)	//RSTOT==65
	{
		for(row=0;row<RSBLOCKS;row++)	//RSBLOCKS==2
		{
			unsigned char c;
			int j;
			c = rsblocks[row][col] ^ m_Scrambler[si++];
			for(j=0;j<8;j++) /* For each bit in a byte */
			{
				sr = (sr<<1) | ((c>>(7-j)) & 1);
				pOutput[w] = parityb(sr & V27POLYB);
				w = update(w);
				pOutput[w] = !parityb(sr & V27POLYA);
				w = update(w);
			}
		}
	}
	//Add convolutional encoder tail
	for(i=0;i<CK-1;i++)
	{
		sr <<= 1;
		pOutput[w] = parityb(sr & V27POLYB);
		w = update(w);
		pOutput[w] = !parityb(sr & V27POLYA);
		w = update(w);
	}
	while(w != -1)
	{
		pOutput[w] = 0;
		w = update(w);
	}
}


////////////////////////////////////////////////////////////////////
//				Copyright 1997 Phil Karn, KA9Q
////////////////////////////////////////////////////////////////////
void CAo40Fec::EncodeRS( unsigned char *data, unsigned char *bb,int pad)
{
int i, j;
unsigned char feedback;

	//Check pad parameter for validity
	if(pad < 0 || pad >= NN)
		return;
	memset(bb,0,NROOTS*sizeof(unsigned char));
	for(i=0;i<NN-NROOTS-pad;i++)
	{
		feedback = m_RS.index_of[data[i] ^ bb[0]];
		if(feedback != NN)      /* feedback term is non-zero */
		{
#ifdef UNNORMALIZED
			// This line is unnecessary when GENPOLY[NROOTS] is unity, as it must
			// always be for the polynomials constructed by init_rs()
			feedback = mod255(NN - CCSDS_poly[NROOTS] + feedback);
#endif
			for(j=1;j<NROOTS;j++)
				bb[j] ^= m_RS.alpha_to[mod255(feedback + m_RS.genpoly[NROOTS-j])];
		}
		// Shift
		memmove(&bb[0],&bb[1],sizeof(unsigned char)*(NROOTS-1));
		if(feedback != NN)
			bb[NROOTS-1] = m_RS.alpha_to[mod255(feedback + m_RS.genpoly[0])];
		else
			bb[NROOTS-1] = 0;
	}
}

//////////////////////////////////////////////////////////////////////////
// Initialize a Reed-Solomon codec
// symsize = symbol size, bits (1-8)
// gfpoly = Field generator polynomial coefficients
// fcr = first root of RS code generator polynomial, index form
// prim = primitive element to generate polynomial roots
// nroots = RS code generator polynomial degree (number of roots)
// pad = padding bytes at front of shortened block
//////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////
//				Copyright 1997 Phil Karn, KA9Q
////////////////////////////////////////////////////////////////////
int CAo40Fec::InitRS(int symsize,int gfpoly,int fcr,int prim, int nroots,int pad)
{
int i, j, sr,root,iprim;

  /* Check parameter ranges */
	if(symsize < 0 || symsize > 8*sizeof(unsigned char))
		return NULL; /* Need version with ints rather than chars */

	if(fcr < 0 || fcr >= (1<<symsize))
		return NULL;
	if(prim <= 0 || prim >= (1<<symsize))
		return NULL;
	if(nroots < 0 || nroots >= (1<<symsize))
		return NULL; /* Can't have more roots than symbol values! */
	if(pad < 0 || pad >= ((1<<symsize) -1 - nroots))
		return NULL; /* Too much padding */

	m_RS.mm = symsize;
	m_RS.nn = (1<<symsize)-1;
	m_RS.pad = pad;

	//deallocate memory if already allocated
	if(m_RS.alpha_to)
		delete m_RS.alpha_to;
	if(m_RS.genpoly)
		delete m_RS.genpoly;
	if(m_RS.index_of)
		delete m_RS.index_of;

	m_RS.alpha_to = new unsigned char[m_RS.nn+1];
	if(m_RS.alpha_to == NULL)
	{
		return NULL;
	}
	m_RS.index_of = new unsigned char[m_RS.nn+1];
	if(m_RS.index_of == NULL)
	{
		delete m_RS.alpha_to;
		return NULL;
	}

	//Generate Galois field lookup tables
	m_RS.index_of[0] = NN; /* log(zero) = -inf */
	m_RS.alpha_to[NN] = 0; /* alpha**-inf = 0 */
	sr = 1;
	for(i=0;i<m_RS.nn;i++)
	{
		m_RS.index_of[sr] = i;
		m_RS.alpha_to[i] = sr;
		sr <<= 1;
		if(sr & (1<<symsize))
			sr ^= gfpoly;
		sr &= m_RS.nn;
	}
	if(sr != 1)
	{  // field generator polynomial is not primitive! 
		delete m_RS.alpha_to;
		delete m_RS.index_of;
		return NULL;
	}

	// Form RS code generator polynomial from its roots 
	m_RS.genpoly = new unsigned char[nroots+1];
	if(m_RS.genpoly == NULL)
	{
		delete m_RS.alpha_to;
		delete m_RS.index_of;
		return NULL;
	}
	m_RS.fcr = fcr;
	m_RS.prim = prim;
	m_RS.nroots = nroots;

	// Find prim-th root of 1, used in decoding 
	for(iprim=1;(iprim % prim) != 0;iprim += m_RS.nn)
	{ }
	m_RS.iprim = iprim / prim;

	m_RS.genpoly[0] = 1;
	for (i = 0,root=fcr*prim; i < nroots; i++,root += prim)
	{
		m_RS.genpoly[i+1] = 1;

	    // Multiply m_RS.genpoly[] by  @**(root + x)
		for (j = i; j > 0; j--)
		{
			if (m_RS.genpoly[j] != 0)
				m_RS.genpoly[j] = m_RS.genpoly[j-1] ^ 
						m_RS.alpha_to[Modnn(m_RS.index_of[m_RS.genpoly[j]] + root)];
			else
				m_RS.genpoly[j] = m_RS.genpoly[j-1];
		}
		// m_RS.genpoly[0] can never be zero
		m_RS.genpoly[0] = m_RS.alpha_to[Modnn(m_RS.index_of[m_RS.genpoly[0]] + root)];
	}
	// convert m_RS.genpoly[] to index form for quicker encoding 
	for (i = 0; i <= nroots; i++)
		m_RS.genpoly[i] = m_RS.index_of[m_RS.genpoly[i]];
	return NULL;
}

//Helper function for mod calc
////////////////////////////////////////////////////////////////////
//				Copyright 1997 Phil Karn, KA9Q
////////////////////////////////////////////////////////////////////
int CAo40Fec::Modnn(int x)
{
	while (x >= m_RS.nn)
	{
		x -= m_RS.nn;
		x = (x >> m_RS.mm) + (x & m_RS.nn);
	}
	return x;
}
