//////////////////////////////////////////////////////////////////////
// RigCntrl.cpp: implementation of the CRigCntrl class.
//
//////////////////////////////////////////////////////////////////////
// Copyright 2000.    Moe Wheatley AE4JY  <ae4jy@qsl.net>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
////////////////////////////////////////////////////////////////////////
//

#include <float.h>
#include <afxmt.h>
#include "stdafx.h"
#include "perform.h"
#include "RigCntrl.h"
#include "BPSK400externals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//Commands for RTS/DTR control
#define COMOFF 0
#define RTSON 1
#define RTSOFF 2
#define DTRON 3
#define DTROFF 4

//Commands for serial commands
#define CMD_OFF 0
#define CMD_UP 1
#define CMD_DN 2

#define PULSE_TIME 25		//mSecs RTS or DTR is high
#define MAX_RATE 50
#define MIN_RATE 0

#define PARAM_MASK_CMDMODE (0xFF000000)
 #define PARAM_RTSDTR 0
 #define PARAM_CMD 1
#define PARAM_MASK_BITRATE (0x00FF0000)
 #define PARAM_1200 0
 #define PARAM_2400 1
 #define PARAM_4800 2
 #define PARAM_9600 3
 #define PARAM_14400 4
 #define PARAM_19200 5
 #define PARAM_38400 6
 #define PARAM_57600 7
#define PARAM_MASK_PARITY (0x0000FF00)
 #define PARAM_NOPARITY 0
 #define PARAM_EVENPARITY 1
 #define PARAM_ODDPARITY 2
#define PARAM_MASK_BITS (0x000000FF)
 #define PARAM_7BITS 0
 #define PARAM_8BITS 1



#define DOPPLER_TIMEOUT 1 //number of .256 seconds of missed blocks before aborting doppler correction

extern double DebugFP1;
extern double DebugFP2;
extern INT DebugINT;


// Non Class routine for launching the worker thread

static UINT ThreadLauncher( CRigCntrl* pIO)
{
	pIO->ProcessLoop();
	return 3;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CRigCntrl::CRigCntrl()
{
	m_fProcThreadQuit = FALSE;
	m_pProcThread = NULL;
	m_ThreadActive = FALSE;
	m_hComPort = INVALID_HANDLE_VALUE;
	m_ComPort = 0;
	m_ComState = 0xFF;
	m_FrequencySetPoint = 1600;
	m_PulseRate = 0;
	m_IntegratorSum = 0.0;
	m_LastDoppler = 0.0;
	m_Mode = RIGCTRL_MODE_MANUAL;
	m_sUp[0] = 'U';
	m_sDn[0] = 'D';
	m_UPCmdLen = 1;
	m_DNCmdLen = 1;
	m_CMDMode = FALSE;
	m_DopplerTimeOut = 0;
	m_StepSize = 10;
	m_BurstCorrectionCount = 0;
	m_BurstCorrectionActive = FALSE;
	QueryPerformanceFrequency( (LARGE_INTEGER*)&m_CountFreq );	//get clock freq
	InitializeCriticalSection(&m_CriticalSection);
}

CRigCntrl::~CRigCntrl()
{
	Stop();
	DeleteResources();
	DeleteCriticalSection(&m_CriticalSection);
}

//*************************************************************************
//**************  routines exposed to the outside  ************************
//*************************************************************************

///////////////////////////////////////////////////////////////////////////
//  Set the port(0 == disable) , mode and other parameters.
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::SetRigCtrlOptions(INT port, INT mode, INT centerfreq,
								  BOOL usb, INT params, INT stepsz)
{
	if( port == 0 )
	{
		Stop();
		return;
	}
	if( ( (params & PARAM_MASK_CMDMODE)>> 24) == PARAM_CMD)
		m_CMDMode = TRUE;
	else
		m_CMDMode = FALSE;
	m_StepSize = stepsz;
	m_ComPort = port;
	m_Mode = mode & ~RIGCTRL_MODE_ACTIVEMASK;
	m_USB = usb;
	m_FrequencySetPoint = centerfreq;
	m_PulseRate = 0;
	m_IntegratorSum = 0.0;
	m_LastDoppler = 0.0;
	m_BurstCorrectionCount = 0;
	m_BurstCorrectionActive = FALSE;
	switch( (params & PARAM_MASK_BITRATE) >> 16 )
	{
		case PARAM_1200: m_BitRate = CBR_1200; break;
		case PARAM_2400: m_BitRate = CBR_2400; break;
		case PARAM_4800: m_BitRate = CBR_4800; break;
		case PARAM_9600: m_BitRate = CBR_9600; break;
		case PARAM_14400: m_BitRate = CBR_14400; break;
		case PARAM_19200: m_BitRate = CBR_19200; break;
		case PARAM_38400: m_BitRate = CBR_38400; break;
		case PARAM_57600: m_BitRate = CBR_57600; break;
	}
	switch( (params & PARAM_MASK_PARITY) >> 8 )
	{
		case PARAM_NOPARITY: m_Parity = NOPARITY; break;
		case PARAM_EVENPARITY: m_Parity = EVENPARITY; break;
		case PARAM_ODDPARITY: m_Parity = ODDPARITY; break;
	}
	switch( (params & PARAM_MASK_BITS)  )
	{
		case PARAM_7BITS: m_NumBits = 7; break;
		case PARAM_8BITS: m_NumBits = 8; break;
	}
	if( (port != 0) && (m_pProcThread== NULL) )
		if( CheckComPort(port) )
			Start();
}

///////////////////////////////////////////////////////////////////////////
//  Set the manual Doppler adjustment rate.
///////////////////////////////////////////////////////////////////////////
INT CRigCntrl::SetManualRate(INT rate)
{
	if( m_Mode == RIGCTRL_MODE_MANUAL)
	{
		m_IntegratorSum = 0.0;
		m_LastDoppler = 0.0;
		CalcPulseTime(rate);
		m_BurstCorrectionActive = FALSE;
		if(m_USB)
			return -m_PulseRate;
		else
			return m_PulseRate;
	}
	else
		return 0;
}

///////////////////////////////////////////////////////////////////////////
//  Set the Command CStrings into character buffer strings. 
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::SetCmdStrings(CString sUp, CString sDn)
{
	m_UPCmdLen = FormatString(sUp, m_sUp);
	m_DNCmdLen = FormatString(sDn, m_sDn);
}

///////////////////////////////////////////////////////////////////////////
// Substitute [xyz] decimal escape sequences into their binary code and place into
// the character buffer.  Returns the new string length.
///////////////////////////////////////////////////////////////////////////
INT CRigCntrl::FormatString(CString inStr, CHAR *pOut)
{
INT sPtr;
INT bPtr = 0;
CHAR ch;
INT len = inStr.GetLength();
	for(sPtr = 0; sPtr<len; sPtr++)
	{
		if(( ( ch=inStr.GetAt(sPtr) ) == '[') && (sPtr<=len-1 ))
		{
			if( ( ch=inStr.GetAt(++sPtr) ) == ']')
			{
				pOut[bPtr++] = '[';	//was just a [] pair
				pOut[bPtr++] = ch;
			}
			else
			{
				pOut[bPtr] = (ch-'0');		//get first digit
				if( (( ch=inStr.GetAt(++sPtr) ) == ']') && (sPtr<len ))
				{
					bPtr++;		//was only 1 digit [x]
				}
				else
				{
					pOut[bPtr] *= 10;
					pOut[bPtr] += (ch-'0');	//get second digit
					if( (( ch=inStr.GetAt(++sPtr) ) == ']') && (sPtr<len ))
					{
						bPtr++;	//was only 2 digits [xy]
					}
					else
					{
						pOut[bPtr] *= 10;
						pOut[bPtr] += (ch-'0');	//get third digit
						if( (( ch=inStr.GetAt(++sPtr) ) == ']') && (sPtr<len ))
						{
							bPtr++;		//got three digits [xyz]
						}
						else
						{
							pOut[bPtr++] = 0;	//error if over 3digits
						}
					}
				}
			}
		}
		else
		{
			pOut[bPtr++] = ch;	//just copy the ACII character
		}
	}
	return bPtr;
}

///////////////////////////////////////////////////////////////////////////
//  Reset the proportional controller variables.
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::ResetCtrl( CBPSKDet* pDet)
{
	m_pBpskdet = pDet;
	if( m_Mode == RIGCTRL_MODE_PROPORTIONAL)
	{
		m_IntegratorSum = 0.0;
		m_LastDoppler = 0.0;
		CalcPulseTime(0);
		m_DopplerTimeOut = 0;
		m_BurstCorrectionCount = 0;
		m_BurstCorrectionActive = FALSE;
	}
}

///////////////////////////////////////////////////////////////////////////
//  Called with current RX frequency sample(Hz), and freqshift rate(Hz/sec)
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::CalculateFrequencyControl( INT freq, double doppler, INT detstatus )
{
#define GAIN_P 0.70
#define GAIN_I 1.0
#define GAIN_IRATE 0.015
#define HYSTER 0

double CFerror;
double Pval=0.0;
double Ival=0.0;
INT error;
	if( m_Mode != RIGCTRL_MODE_PROPORTIONAL)
		return;

	if( detstatus == (DLLSTAT_RXCRC_NO_SIGNAL|DLLSTAT_RXFEC_NO_SIGNAL) )
	{
		m_BurstCorrectionCount = 0;
		m_BurstCorrectionActive = FALSE;
		if( ++m_DopplerTimeOut >= DOPPLER_TIMEOUT)
		{
			ResetCtrl(m_pBpskdet);
		}
		return;
	}
	m_DopplerTimeOut = 0;

// Get position error from desired center frequency
	if( m_StepSize > 20 )
	{
		CalcPulseTime(0);
		if( DLLSTAT_RXCRC_SYNC_SEARCH == (detstatus&DLLSTAT_RXCRC_MASK)  )
		{
			if(!m_BurstCorrectionActive)
			{
				error = m_FrequencySetPoint - freq;
				if(m_USB)
					error = -error;
				m_BurstCorrectionCount = error/m_StepSize;
				m_PulseRate = m_BurstCorrectionCount*m_StepSize;
				m_BurstCorrectionActive = TRUE;
				if(m_BurstCorrectionCount > 0 )
					m_AdjStepSize = m_StepSize;
				else
					m_AdjStepSize = -m_StepSize;
				if(m_USB)
					m_AdjStepSize = -m_AdjStepSize;
			}
		}
		else
		{
			m_BurstCorrectionCount = 0;
			m_BurstCorrectionActive = FALSE;
		}
	}
	else
	{
		CFerror = (double)(m_FrequencySetPoint - freq);
		if(CFerror > HYSTER )
		{
			if(doppler < CFerror) 
				error = (INT)(CFerror);
			else
				error = 0;	//just let doppler drift the signal in
		}
		else 
		{
			if(CFerror < -HYSTER )
			{
				if(doppler >= CFerror) 
					error = (INT)(CFerror);
				else
					error = 0;	//just let doppler drift the signal in
			}
			else
				error = 0;
		}
//Simple PI control loop
		Pval = GAIN_P*error;	//proportional component 
		m_IntegratorSum += (GAIN_IRATE*error);
		if(m_IntegratorSum > MAX_RATE)
			m_IntegratorSum = MAX_RATE;
		if(m_IntegratorSum < -MAX_RATE)
			m_IntegratorSum = -MAX_RATE;
		Ival = m_IntegratorSum*GAIN_I;	//Integral component 
//DebugFP1= Pval;
//DebugFP2= m_IntegratorSum;
		CalcPulseTime((INT)(Pval+Ival));	//update output rate generator
	}
}

///////////////////////////////////////////////////////////////////////////
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::Stop()
{
	if( m_pProcThread != NULL )		//if thread is already running
	{
		m_ThreadActive = FALSE;
		m_fProcThreadQuit = TRUE;	// stop it
		::WaitForSingleObject(m_pProcThread->m_hThread,	2500);
		delete m_pProcThread;
		m_pProcThread = NULL;
		m_hComPort = INVALID_HANDLE_VALUE;
	}
}


///////////////////////////////////////////////////////////////////////////
//   Called to start the I/O thread running
///////////////////////////////////////////////////////////////////////////
INT CRigCntrl::Start()
{
	if( m_pProcThread == NULL)
	{
		m_ThreadActive = FALSE;
		m_fProcThreadQuit = FALSE;
		m_hComPort = INVALID_HANDLE_VALUE;
		m_pProcThread = AfxBeginThread(
			(AFX_THREADPROC)ThreadLauncher,	//thread function
			(LPVOID)this,					// ptr to this class
			THREAD_PRIORITY_LOWEST, 		//give worker thread priority
			0,								// same stack size
			CREATE_SUSPENDED,				// don't let it start yet
			NULL );							// same security attributes
		if(!m_pProcThread)
			return TRUE;
		m_pProcThread->m_bAutoDelete = FALSE;	// keep thread object around
		m_pProcThread->ResumeThread ();		//let er rip
	}
	return FALSE;
}


///////////////////////////////////////////////////////////////////////////
//call to check for the existence of com 'port'
// Returns TRUE if exists or port==0
///////////////////////////////////////////////////////////////////////////
BOOL CRigCntrl::CheckComPort(LONG port)
{
char szPort[10];
BOOL ret = FALSE;
HANDLE hComPort;
	Stop();
	if( (port == 0) || (port == m_ComPort) )	//if already in use
		return TRUE;
	wsprintf( szPort, "COM%d", port );
	EnterCriticalSection(&m_CriticalSection);
	hComPort = CreateFile( szPort, 0, 0, NULL,
						OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if( hComPort == INVALID_HANDLE_VALUE )
	{
		CloseHandle( hComPort );
		LeaveCriticalSection(&m_CriticalSection);
		return FALSE;
	}
	else
	{
		CloseHandle( hComPort );
		LeaveCriticalSection(&m_CriticalSection);
		return TRUE;
	}
}

//*************************************************************************
//**********************  local routines  *********************************
//*************************************************************************

///////////////////////////////////////////////////////////////////////////
//  Main I/O thread processing loop
// Process either pulsing the RTS/DTR lines or sending serial cmds
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::ProcessLoop()
{
	m_ThreadActive = TRUE;
	m_ComState = 0xFF;
	if( m_CMDMode )
		ProcessCMDLoop();
	else
		ProcessRTSDTRLoop();
	DeleteResources();		//kill all resources
}

///////////////////////////////////////////////////////////////////////////
// Process sending serial up dn cmds
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::ProcessCMDLoop()
{
LONGLONG DeltaTime = 0;
INT pulses;
	while( !m_fProcThreadQuit )
	{
		if( m_BurstCorrectionCount != 0 )
		{
			pulses = abs(m_BurstCorrectionCount);
			for( INT i=0; i<pulses; i++)
			{
				if( m_BurstCorrectionCount > 0 )
					WriteCMDString( m_sUp,m_UPCmdLen);
				else
					WriteCMDString( m_sDn,m_DNCmdLen);
				m_pBpskdet->AdjDopplerJmp(m_AdjStepSize);
				Sleep(200);
			}
			m_BurstCorrectionCount = 0;
		}
		else
		{
			if( m_PulseRate == 0 )
			{
				CloseComPort();
				Sleep(PULSE_TIME*2);
			}
			else
			{
	//StartPerformance();
				QueryPerformanceCounter( (LARGE_INTEGER*)&m_StartTime );
				if( m_PulseRate > 0 )
				{
					WriteCMDString( m_sUp,m_UPCmdLen);
				}
				else
				{
					WriteCMDString( m_sDn,m_DNCmdLen);
				}
				if( m_OffTime>250 )
					CloseComPort();
				do{
					QueryPerformanceCounter( (LARGE_INTEGER*)&m_StopTime );
					DeltaTime = m_StopTime - m_StartTime;
					DeltaTime = (DeltaTime*1000)/m_CountFreq;
					Sleep(m_OffTime/100);
				}while( (DeltaTime < (LONGLONG)(m_OffTime+PULSE_TIME) ) && !m_fProcThreadQuit );
//StopPerformance();
			}
		}
	}
	CloseComPort();
}

///////////////////////////////////////////////////////////////////////////
// Process pulsing the RTS/DTR lines
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::ProcessRTSDTRLoop()
{
LONGLONG DeltaTime = 0;
INT pulses;
	while( !m_fProcThreadQuit )
	{
		if( m_BurstCorrectionCount != 0 )
		{
			pulses = abs(m_BurstCorrectionCount);
			for( INT i=0; i<pulses; i++)
			{
				QueryPerformanceCounter( (LARGE_INTEGER*)&m_StartTime );
				if( m_BurstCorrectionCount > 0 )
					RTSDTRCntrl(RTSON);		//positive rate pulse
				else
					RTSDTRCntrl(DTRON);		//negative rate pulse
				do{
					QueryPerformanceCounter( (LARGE_INTEGER*)&m_StopTime );
					DeltaTime = m_StopTime - m_StartTime;
					DeltaTime = (DeltaTime*1000)/m_CountFreq;
					Sleep(0);
				}while( (DeltaTime < (LONGLONG)PULSE_TIME) && !m_fProcThreadQuit );
				RTSDTRCntrl(RTSOFF);
				RTSDTRCntrl(DTROFF);
				m_pBpskdet->AdjDopplerJmp(m_AdjStepSize);
				Sleep(200);
			}
			m_BurstCorrectionCount = 0;
		}
		else
		{
			if( m_PulseRate == 0 )
			{
				RTSDTRCntrl(RTSOFF);
				RTSDTRCntrl(DTROFF);
				Sleep(PULSE_TIME*2);
			}
			else
			{
				QueryPerformanceCounter( (LARGE_INTEGER*)&m_StartTime );
				if( m_PulseRate > 0 )
				{
					RTSDTRCntrl(RTSON);		//positive rate pulse
					do{
						QueryPerformanceCounter( (LARGE_INTEGER*)&m_StopTime );
						DeltaTime = m_StopTime - m_StartTime;
						DeltaTime = (DeltaTime*1000)/m_CountFreq;
						Sleep(0);
					}while( (DeltaTime < (LONGLONG)PULSE_TIME) && !m_fProcThreadQuit );
					RTSDTRCntrl(RTSOFF);
				}
				else
				{
					RTSDTRCntrl(DTRON);		//negative rate pulse
					do{
						QueryPerformanceCounter( (LARGE_INTEGER*)&m_StopTime );
						DeltaTime = m_StopTime - m_StartTime;
						DeltaTime = (DeltaTime*1000)/m_CountFreq;
						Sleep(0);
					}while( (DeltaTime < (LONGLONG)PULSE_TIME) && !m_fProcThreadQuit );
					RTSDTRCntrl(DTROFF);
				}
				do{
					QueryPerformanceCounter( (LARGE_INTEGER*)&m_StopTime );
					DeltaTime = m_StopTime - m_StartTime;
					DeltaTime = (DeltaTime*1000)/m_CountFreq;
					Sleep(m_OffTime/100);
				}while( (DeltaTime < (LONGLONG)(m_OffTime+PULSE_TIME) ) && !m_fProcThreadQuit );
			}
		}
	}
}


///////////////////////////////////////////////////////////////////////////
//  Free resources and exit process thread
//   Called by process thread itself to shut itself off if not signaled by
//  main program thread.
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::StopThread()
{
//  if any resources are open, kill em.
	DeleteResources();
	m_pProcThread->m_bAutoDelete = TRUE;	//make thread clean up after
	m_pProcThread = NULL;  				// itself before it dies
	AfxEndThread(4);
}

///////////////////////////////////////////////////////////////////////////
//   Called to delete any allocated resources in this class
///////////////////////////////////////////////////////////////////////////
void CRigCntrl::DeleteResources()
{
	m_PulseRate = 0;
	RTSDTRCntrl(RTSOFF);
	RTSDTRCntrl(DTROFF);
	RTSDTRCntrl(COMOFF);
	Sleep(10);
	m_ThreadActive = FALSE;
}



//////////////////////////////////////////////////////////////////
//  Calculate pulse off time based on the pulse width and rate.
//////////////////////////////////////////////////////////////////
void CRigCntrl::CalcPulseTime(INT rate)
{
	if( rate > MAX_RATE)
		rate = MAX_RATE;
	else
		if( rate < -MAX_RATE)
			rate = -MAX_RATE;
	if(m_USB)
		rate = -rate;
	EnterCriticalSection(&m_CriticalSection);
	m_PulseRate = rate;
	if(rate != 0)
		m_OffTime = (m_StepSize*1000)/abs(rate) - PULSE_TIME;
	else
		m_OffTime = PULSE_TIME*2;
	LeaveCriticalSection(&m_CriticalSection);
}

//////////////////////////////////////////////////////////////////
//  Controls COM port signals DTR and RTS for freq step control
// possible states are COMOFF RTSON RTSOFF DTRON DTROFF
//////////////////////////////////////////////////////////////////
void CRigCntrl::RTSDTRCntrl(INT state)
{
char szPort[10];
	if( m_ComPort != 0 )
	{
		if(state != m_ComState)	//if is changed
		{
			m_ComState = state;
			if( m_hComPort == INVALID_HANDLE_VALUE ) //need to open?
			{
				wsprintf( szPort, "COM%d", m_ComPort );
				EnterCriticalSection(&m_CriticalSection);
				m_hComPort = CreateFile( szPort, GENERIC_READ | GENERIC_WRITE, 0, NULL,
								OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
				LeaveCriticalSection(&m_CriticalSection);
				if( m_hComPort == INVALID_HANDLE_VALUE )
					return;
			}
			switch(m_ComState)
			{
				case RTSON:
					EscapeCommFunction(m_hComPort, SETRTS);
					EscapeCommFunction(m_hComPort, CLRDTR);
					break;
				case RTSOFF:
					EscapeCommFunction(m_hComPort, CLRRTS);
					EscapeCommFunction(m_hComPort, CLRDTR);
					break;
				case DTRON:
					EscapeCommFunction(m_hComPort, SETDTR);
					EscapeCommFunction(m_hComPort, CLRRTS);
					break;
				case DTROFF:
					EscapeCommFunction(m_hComPort, CLRRTS);
					EscapeCommFunction(m_hComPort, CLRDTR);
					break;
				case COMOFF:
					EscapeCommFunction(m_hComPort, CLRRTS);
					EscapeCommFunction(m_hComPort, CLRDTR);
					CloseComPort();
					break;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////
//  Opens serial COM port for command sending mode
//////////////////////////////////////////////////////////////////
void CRigCntrl::OpenComPort()
{
char szPort[10];
DCB dcb;
	if( m_hComPort == INVALID_HANDLE_VALUE ) //need to open?
	{
		wsprintf( szPort, "COM%d", m_ComPort );
		EnterCriticalSection(&m_CriticalSection);
		m_hComPort = CreateFile( szPort, GENERIC_READ | GENERIC_WRITE, 0, NULL,
						OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
		LeaveCriticalSection(&m_CriticalSection);
		if( m_hComPort == INVALID_HANDLE_VALUE )
			return;
	}
	else
		return;
	if(!GetCommState(m_hComPort, &dcb) )
	{
		CloseComPort();
		return;
	}
	dcb.fDsrSensitivity = FALSE;
	dcb.BaudRate = m_BitRate;
	dcb.ByteSize = m_NumBits;
	dcb.Parity = m_Parity;
	dcb.StopBits = TWOSTOPBITS;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	if(!SetCommState(m_hComPort, &dcb))
	{
		CloseComPort();
		return;
	}

}

//////////////////////////////////////////////////////////////////
//  Closes serial COM port if open
//////////////////////////////////////////////////////////////////
void CRigCntrl::CloseComPort()
{
	if(m_hComPort != INVALID_HANDLE_VALUE)
	{
		CloseHandle( m_hComPort );
		m_hComPort = INVALID_HANDLE_VALUE;
	}
}

//////////////////////////////////////////////////////////////////
//  Writes serial command string of length 'length'
//////////////////////////////////////////////////////////////////
void CRigCntrl::WriteCMDString(const char *str, INT length)
{
DWORD dwBytesWritten = 0;
	OpenComPort();
	if(m_hComPort != INVALID_HANDLE_VALUE)
	{
		if(!WriteFile(m_hComPort, str, length, &dwBytesWritten,NULL) )
		{
			CloseComPort();
		}
		if(dwBytesWritten != (DWORD)length)
		{
			CloseComPort();
		}
	}
}

